/*
 * Jalview - A Sequence Alignment Editor and Viewer ($$Version-Rel$$)
 * Copyright (C) $$Year-Rel$$ The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */
package jalview.ext.rbvi.chimera;

import jalview.api.AlignViewportI;
import jalview.api.AlignmentViewPanel;
import jalview.api.FeatureRenderer;
import jalview.datamodel.AlignmentI;
import jalview.datamodel.MappedFeatures;
import jalview.datamodel.SequenceFeature;
import jalview.datamodel.SequenceI;
import jalview.gui.Desktop;
import jalview.structure.AtomSpecModel;
import jalview.structure.StructureCommand;
import jalview.structure.StructureCommandI;
import jalview.structure.StructureCommandsBase;
import jalview.structure.StructureMapping;
import jalview.structure.StructureSelectionManager;
import jalview.util.ColorUtils;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

/**
 * Routines for generating Chimera commands for Jalview/Chimera binding
 * 
 * @author JimP
 * 
 */
public class ChimeraCommands extends StructureCommandsBase
{
  private static final StructureCommand SHOW_BACKBONE = new StructureCommand(
          "~display all;~ribbon;chain @CA|P");

  public static final String NAMESPACE_PREFIX = "jv_";

  private static final StructureCommandI COLOUR_BY_CHARGE = new StructureCommand(
          "color white;color red ::ASP,GLU;color blue ::LYS,ARG;color yellow ::CYS");

  private static final StructureCommandI COLOUR_BY_CHAIN = new StructureCommand(
          "rainbow chain");

  // Chimera clause to exclude alternate locations in atom selection
  private static final String NO_ALTLOCS = "&~@.B-Z&~@.2-9";

  @Override
  public StructureCommandI getColourCommand(String atomSpec, Color colour)
  {
    // https://www.cgl.ucsf.edu/chimera/current/docs/UsersGuide/midas/color.html
    String colourCode = getColourString(colour);
    return new StructureCommand("color " + colourCode + " " + atomSpec);
  }

  /**
   * Returns a colour formatted suitable for use in viewer command syntax
   * 
   * @param colour
   * @return
   */
  protected String getColourString(Color colour)
  {
    return ColorUtils.toTkCode(colour);
  }

  /**
   * Constructs and returns Chimera commands to set attributes on residues
   * corresponding to features in Jalview. Attribute names are the Jalview feature
   * type, with a "jv_" prefix.
   * 
   * @param ssm
   * @param files
   * @param seqs
   * @param viewPanel
   * @return
   */
  @Override
  public List<StructureCommandI> setAttributesForFeatures(
          StructureSelectionManager ssm, String[] files, SequenceI[][] seqs,
          AlignmentViewPanel viewPanel)
  {
    Map<String, Map<Object, AtomSpecModel>> featureMap = buildFeaturesMap(
            ssm, files, seqs, viewPanel);

    return setAttributes(featureMap);
  }

  /**
   * <pre>
   * Helper method to build a map of 
   *   { featureType, { feature value, AtomSpecModel } }
   * </pre>
   * 
   * @param ssm
   * @param files
   * @param seqs
   * @param viewPanel
   * @return
   */
  protected Map<String, Map<Object, AtomSpecModel>> buildFeaturesMap(
          StructureSelectionManager ssm, String[] files, SequenceI[][] seqs,
          AlignmentViewPanel viewPanel)
  {
    Map<String, Map<Object, AtomSpecModel>> theMap = new LinkedHashMap<>();

    FeatureRenderer fr = viewPanel.getFeatureRenderer();
    if (fr == null)
    {
      return theMap;
    }

    AlignViewportI viewport = viewPanel.getAlignViewport();
    List<String> visibleFeatures = fr.getDisplayedFeatureTypes();

    /*
     * if alignment is showing features from complement, we also transfer
     * these features to the corresponding mapped structure residues
     */
    boolean showLinkedFeatures = viewport.isShowComplementFeatures();
    List<String> complementFeatures = new ArrayList<>();
    FeatureRenderer complementRenderer = null;
    if (showLinkedFeatures)
    {
      AlignViewportI comp = fr.getViewport().getCodingComplement();
      if (comp != null)
      {
        complementRenderer = Desktop.getAlignFrameFor(comp)
                .getFeatureRenderer();
        complementFeatures = complementRenderer.getDisplayedFeatureTypes();
      }
    }
    if (visibleFeatures.isEmpty() && complementFeatures.isEmpty())
    {
      return theMap;
    }

    AlignmentI alignment = viewPanel.getAlignment();
    for (int pdbfnum = 0; pdbfnum < files.length; pdbfnum++)
    {
      final int modelNumber = pdbfnum + getModelStartNo();
      String modelId = String.valueOf(modelNumber);
      StructureMapping[] mapping = ssm.getMapping(files[pdbfnum]);

      if (mapping == null || mapping.length < 1)
      {
        continue;
      }

      for (int seqNo = 0; seqNo < seqs[pdbfnum].length; seqNo++)
      {
        for (int m = 0; m < mapping.length; m++)
        {
          final SequenceI seq = seqs[pdbfnum][seqNo];
          int sp = alignment.findIndex(seq);
          StructureMapping structureMapping = mapping[m];
          if (structureMapping.getSequence() == seq && sp > -1)
          {
            /*
             * found a sequence with a mapping to a structure;
             * now scan its features
             */
            if (!visibleFeatures.isEmpty())
            {
              scanSequenceFeatures(visibleFeatures, structureMapping, seq,
                      theMap, modelId);
            }
            if (showLinkedFeatures)
            {
              scanComplementFeatures(complementRenderer, structureMapping,
                      seq, theMap, modelId);
            }
          }
        }
      }
    }
    return theMap;
  }

  /**
   * Scans visible features in mapped positions of the CDS/peptide complement, and
   * adds any found to the map of attribute values/structure positions
   * 
   * @param complementRenderer
   * @param structureMapping
   * @param seq
   * @param theMap
   * @param modelNumber
   */
  protected static void scanComplementFeatures(
          FeatureRenderer complementRenderer,
          StructureMapping structureMapping, SequenceI seq,
          Map<String, Map<Object, AtomSpecModel>> theMap,
          String modelNumber)
  {
    /*
     * for each sequence residue mapped to a structure position...
     */
    for (int seqPos : structureMapping.getMapping().keySet())
    {
      /*
       * find visible complementary features at mapped position(s)
       */
      MappedFeatures mf = complementRenderer
              .findComplementFeaturesAtResidue(seq, seqPos);
      if (mf != null)
      {
        for (SequenceFeature sf : mf.features)
        {
          String type = sf.getType();

          /*
           * Don't copy features which originated from Chimera
           */
          if (JalviewChimeraBinding.CHIMERA_FEATURE_GROUP
                  .equals(sf.getFeatureGroup()))
          {
            continue;
          }

          /*
           * record feature 'value' (score/description/type) as at the
           * corresponding structure position
           */
          List<int[]> mappedRanges = structureMapping
                  .getPDBResNumRanges(seqPos, seqPos);

          if (!mappedRanges.isEmpty())
          {
            String value = sf.getDescription();
            if (value == null || value.length() == 0)
            {
              value = type;
            }
            float score = sf.getScore();
            if (score != 0f && !Float.isNaN(score))
            {
              value = Float.toString(score);
            }
            Map<Object, AtomSpecModel> featureValues = theMap.get(type);
            if (featureValues == null)
            {
              featureValues = new HashMap<>();
              theMap.put(type, featureValues);
            }
            for (int[] range : mappedRanges)
            {
              addAtomSpecRange(featureValues, value, modelNumber, range[0],
                      range[1], structureMapping.getChain());
            }
          }
        }
      }
    }
  }

  /**
   * Inspect features on the sequence; for each feature that is visible,
   * determine its mapped ranges in the structure (if any) according to the
   * given mapping, and add them to the map.
   * 
   * @param visibleFeatures
   * @param mapping
   * @param seq
   * @param theMap
   * @param modelId
   */
  protected static void scanSequenceFeatures(List<String> visibleFeatures,
          StructureMapping mapping, SequenceI seq,
          Map<String, Map<Object, AtomSpecModel>> theMap, String modelId)
  {
    List<SequenceFeature> sfs = seq.getFeatures().getPositionalFeatures(
            visibleFeatures.toArray(new String[visibleFeatures.size()]));
    for (SequenceFeature sf : sfs)
    {
      String type = sf.getType();

      /*
       * Don't copy features which originated from Chimera
       */
      if (JalviewChimeraBinding.CHIMERA_FEATURE_GROUP
              .equals(sf.getFeatureGroup()))
      {
        continue;
      }

      List<int[]> mappedRanges = mapping.getPDBResNumRanges(sf.getBegin(),
              sf.getEnd());

      if (!mappedRanges.isEmpty())
      {
        String value = sf.getDescription();
        if (value == null || value.length() == 0)
        {
          value = type;
        }
        float score = sf.getScore();
        if (score != 0f && !Float.isNaN(score))
        {
          value = Float.toString(score);
        }
        Map<Object, AtomSpecModel> featureValues = theMap.get(type);
        if (featureValues == null)
        {
          featureValues = new HashMap<>();
          theMap.put(type, featureValues);
        }
        for (int[] range : mappedRanges)
        {
          addAtomSpecRange(featureValues, value, modelId, range[0],
                  range[1], mapping.getChain());
        }
      }
    }
  }

  /**
   * Traverse the map of features/values/models/chains/positions to construct a
   * list of 'setattr' commands (one per distinct feature type and value).
   * <p>
   * The format of each command is
   * 
   * <pre>
   * <blockquote> setattr r <featureName> " " #modelnumber:range.chain 
   * e.g. setattr r jv_chain &lt;value&gt; #0:2.B,4.B,9-12.B|#1:1.A,2-6.A,...
   * </blockquote>
   * </pre>
   * 
   * @param featureMap
   * @return
   */
  protected List<StructureCommandI> setAttributes(
          Map<String, Map<Object, AtomSpecModel>> featureMap)
  {
    List<StructureCommandI> commands = new ArrayList<>();
    for (String featureType : featureMap.keySet())
    {
      String attributeName = makeAttributeName(featureType);

      /*
       * clear down existing attributes for this feature
       */
      // 'problem' - sets attribute to None on all residues - overkill?
      // commands.add("~setattr r " + attributeName + " :*");

      Map<Object, AtomSpecModel> values = featureMap.get(featureType);
      for (Object value : values.keySet())
      {
        /*
         * for each distinct value recorded for this feature type,
         * add a command to set the attribute on the mapped residues
         * Put values in single quotes, encoding any embedded single quotes
         */
        AtomSpecModel atomSpecModel = values.get(value);
        String featureValue = value.toString();
        featureValue = featureValue.replaceAll("\\'", "&#39;");
        StructureCommandI cmd = setAttribute(attributeName, featureValue,
                atomSpecModel);
        commands.add(cmd);
      }
    }

    return commands;
  }

  /**
   * Returns a viewer command to set the given residue attribute value on
   * residues specified by the AtomSpecModel, for example
   * 
   * <pre>
   * setatr res jv_chain 'primary' #1:12-34,48-55.B
   * </pre>
   * 
   * @param attributeName
   * @param attributeValue
   * @param atomSpecModel
   * @return
   */
  protected StructureCommandI setAttribute(String attributeName,
          String attributeValue,
          AtomSpecModel atomSpecModel)
  {
    StringBuilder sb = new StringBuilder(128);
    sb.append("setattr res ").append(attributeName).append(" '")
            .append(attributeValue).append("' ");
    sb.append(getAtomSpec(atomSpecModel, false));
    return new StructureCommand(sb.toString());
  }

  /**
   * Makes a prefixed and valid Chimera attribute name. A jv_ prefix is applied
   * for a 'Jalview' namespace, and any non-alphanumeric character is converted
   * to an underscore.
   * 
   * @param featureType
   * @return
   * @see https://www.cgl.ucsf.edu/chimera/current/docs/UsersGuide/midas/setattr.html
   */
  protected static String makeAttributeName(String featureType)
  {
    StringBuilder sb = new StringBuilder();
    if (featureType != null)
    {
      for (char c : featureType.toCharArray())
      {
        sb.append(Character.isLetterOrDigit(c) ? c : '_');
      }
    }
    String attName = NAMESPACE_PREFIX + sb.toString();

    /*
     * Chimera treats an attribute name ending in 'color' as colour-valued;
     * Jalview doesn't, so prevent this by appending an underscore
     */
    if (attName.toUpperCase().endsWith("COLOR"))
    {
      attName += "_";
    }

    return attName;
  }

  @Override
  public StructureCommandI colourByChain()
  {
    return COLOUR_BY_CHAIN;
  }

  @Override
  public List<StructureCommandI> colourByCharge()
  {
    return Arrays.asList(COLOUR_BY_CHARGE);
  }

  @Override
  public String getResidueSpec(String residue)
  {
    return "::" + residue;
  }

  @Override
  public StructureCommandI setBackgroundColour(Color col)
  {
    // https://www.cgl.ucsf.edu/chimera/current/docs/UsersGuide/midas/set.html#bgcolor
    return new StructureCommand("set bgColor " + ColorUtils.toTkCode(col));
  }

  @Override
  public StructureCommandI focusView()
  {
    // https://www.cgl.ucsf.edu/chimera/current/docs/UsersGuide/midas/focus.html
    return new StructureCommand("focus");
  }

  @Override
  public List<StructureCommandI> showChains(List<String> toShow)
  {
    /*
     * Construct a chimera command like
     * 
     * ~display #*;~ribbon #*;ribbon :.A,:.B
     */
    StringBuilder cmd = new StringBuilder(64);
    boolean first = true;
    for (String chain : toShow)
    {
      String[] tokens = chain.split(":");
      if (tokens.length == 2)
      {
        String showChainCmd = tokens[0] + ":." + tokens[1];
        if (!first)
        {
          cmd.append(",");
        }
        cmd.append(showChainCmd);
        first = false;
      }
    }

    /*
     * could append ";focus" to this command to resize the display to fill the
     * window, but it looks more helpful not to (easier to relate chains to the
     * whole)
     */
    final String command = "~display #*; ~ribbon #*; ribbon :"
            + cmd.toString();
    return Arrays.asList(new StructureCommand(command));
  }

  @Override
  public List<StructureCommandI> superposeStructures(AtomSpecModel ref,
          AtomSpecModel spec)
  {
    /*
     * Form Chimera match command to match spec to ref
     * (the first set of atoms are moved on to the second)
     * 
     * match #1:1-30.B,81-100.B@CA #0:21-40.A,61-90.A@CA
     * 
     * @see https://www.cgl.ucsf.edu/chimera/docs/UsersGuide/midas/match.html
     */
    StringBuilder cmd = new StringBuilder();
    String atomSpecAlphaOnly = getAtomSpec(spec, true);
    String refSpecAlphaOnly = getAtomSpec(ref, true);
    cmd.append("match ").append(atomSpecAlphaOnly).append(" ").append(refSpecAlphaOnly);

    /*
     * show superposed residues as ribbon
     */
    String atomSpec = getAtomSpec(spec, false);
    String refSpec = getAtomSpec(ref, false);
    cmd.append("; ribbon ");
    cmd.append(atomSpec).append("|").append(refSpec).append("; focus");

    return Arrays.asList(new StructureCommand(cmd.toString()));
  }

  @Override
  public StructureCommandI openCommandFile(String path)
  {
    // https://www.cgl.ucsf.edu/chimera/current/docs/UsersGuide/filetypes.html
    return new StructureCommand("open cmd:" + path);
  }

  @Override
  public StructureCommandI saveSession(String filepath)
  {
    // https://www.cgl.ucsf.edu/chimera/current/docs/UsersGuide/midas/save.html
    return new StructureCommand("save " + filepath);
  }

  /**
   * Returns the range(s) modelled by {@code atomSpec} formatted as a Chimera
   * atomspec string, e.g.
   * 
   * <pre>
   * #0:15.A,28.A,54.A,70-72.A|#1:2.A,6.A,11.A,13-14.A
   * </pre>
   * 
   * where
   * <ul>
   * <li>#0 is a model number</li>
   * <li>15 or 70-72 is a residue number, or range of residue numbers</li>
   * <li>.A is a chain identifier</li>
   * <li>residue ranges are separated by comma</li>
   * <li>atomspecs for distinct models are separated by | (or)</li>
   * </ul>
   * 
   * <pre>
   * 
   * @param model
   * @param alphaOnly
   * @return
   * @see https://www.cgl.ucsf.edu/chimera/current/docs/UsersGuide/midas/frameatom_spec.html
   */
  @Override
  public String getAtomSpec(AtomSpecModel atomSpec, boolean alphaOnly)
  {
    StringBuilder sb = new StringBuilder(128);
    boolean firstModel = true;
    for (String model : atomSpec.getModels())
    {
      if (!firstModel)
      {
        sb.append("|");
      }
      firstModel = false;
      appendModel(sb, model, atomSpec, alphaOnly);
    }
    return sb.toString();
  }

  /**
   * A helper method to append an atomSpec string for atoms in the given model
   * 
   * @param sb
   * @param model
   * @param atomSpec
   * @param alphaOnly
   */
  protected void appendModel(StringBuilder sb, String model,
          AtomSpecModel atomSpec, boolean alphaOnly)
  {
    sb.append("#").append(model).append(":");

    boolean firstPositionForModel = true;

    for (String chain : atomSpec.getChains(model))
    {
      chain = " ".equals(chain) ? chain : chain.trim();

      List<int[]> rangeList = atomSpec.getRanges(model, chain);
      for (int[] range : rangeList)
      {
        appendRange(sb, range[0], range[1], chain, firstPositionForModel,
                false);
        firstPositionForModel = false;
      }
    }
    if (alphaOnly)
    {
      /*
       * restrict to alpha carbon, no alternative locations
       * (needed to ensuring matching atom counts for superposition)
       */
      // TODO @P instead if RNA - add nucleotide flag to AtomSpecModel?
      sb.append("@CA").append(NO_ALTLOCS);
    }
  }

  @Override
  public List<StructureCommandI> showBackbone()
  {
    return Arrays.asList(SHOW_BACKBONE);
  }

  @Override
  public StructureCommandI loadFile(String file)
  {
    return new StructureCommand("open " + file);
  }

}
